from AnkiIn.note import Note
from AnkiIn.notetypes.QA import get as super_get
from AnkiIn.notetypes.QA import check as super_check
from AnkiIn.model import Model
from AnkiIn.config import dict as conf
from AnkiIn.config import config_updater
from AnkiIn.notetypes.QA import QANote


notetype_name = "SQA"
if notetype_name not in conf["notetype"]:
    conf["notetype"][notetype_name] = {}
settings = conf["notetype"][notetype_name]

priority = None


def update_siyuan_qa_config():
    global settings, priority

    priority = settings.get("priority", 11)


config_updater.append((update_siyuan_qa_config, 10))


def check(lines: list, extra_params={}) -> bool:
    return super_check(lines=lines, extra_params=extra_params)


def get(text: str, deck: str, tags: list, extra_params={}) -> Note:
    return SQANote(extra_params["SiyuanID"], super_get(text=text, deck=deck, tags=tags, extra_params=extra_params))


BACK = r"""{{FrontSide}}
<hr id=answer>
{{Back}}
</br>
<a href="siyuan://blocks/{{SiyuanID}}">Open in SiYuan</a>"""

CSS = r""".card {
  font-family: arial;
  font-size: 20px;
  text-align: center;
  color: black;
  background-color: white;
}
ul {
display: inline-block;
text-align: left;
}
ol {
display: inline-block;
text-align: left;
}
"""

MODELNAME = "SAnkiLink-Basic"
MODELID = 1145841921

_model = Model(
    modelId=MODELID,
    modelName=MODELNAME,
    fields=["Front", "Back", "SiyuanID"],
    templates=[
        {
            'Name': 'Card 1',
            'Front': '{{Front}}',
            'Back': BACK
        }
    ],
    css=CSS
)


class SQANote(QANote):
    def __init__(self, SiyuanID, front, back, deck, tags):
        global _model
        super.__init__(front=front, back=back, deck=deck, tags=tags)
        self.model = _model
        self["SiyuanID"] = SiyuanID

    def __init__(self, SiyuanID, Note):
        global _model
        self.__dict__.update(Note.__dict__)
        self.model = _model
        self["SiyuanID"] = SiyuanID



from AnkiIn.note import Note
from AnkiIn.model import Model
from AnkiIn.config import dict as conf
from AnkiIn.config import config_updater
from AnkiIn.log import notetype_logger as log


notetype_name = "SWatch"
if notetype_name not in conf["notetype"]:
    conf["notetype"][notetype_name] = {}
settings = conf["notetype"][notetype_name]

priority = None
enable = None


def update_swatch_config():
    global settings, priority, enable

    priority = settings.get("priority", 1)
    enable = settings.get("enable", False)


config_updater.append((update_swatch_config, 10))


def check(lines: list, extra_params={}) -> bool:
    return enable and len(lines) >= 1


def get(text: str, deck: str, tags: list, extra_params={}) -> Note:
    return SWatchNote(front=text, SiyuanID=extra_params["SiyuanID"], deck=deck, tags=tags)


BACK = r"""{{FrontSide}}
<hr id=answer>
<a href="siyuan://blocks/{{SiyuanID}}">Open in SiYuan</a>
"""

CSS = r""".card {
  font-family: arial;
  font-size: 20px;
  text-align: center;
  color: black;
  background-color: white;
}
ul {
display: inline-block;
text-align: left;
}
ol {
display: inline-block;
text-align: left;
}
"""

MODELNAME = "SAnkiLink-Watch"
MODELID = 1195191921

_model = Model(
    modelId=MODELID,
    modelName=MODELNAME,
    fields=["Front", "SiyuanID"],
    templates=[
        {
            'Name': 'Card 1',
            'Front': '{{Front}}',
            'Back': BACK
        }
    ],
    css=CSS
)


class SWatchNote(Note):
    def __init__(self, front, SiyuanID, deck, tags):
        global _model
        super().__init__(model=_model, fields={
            "Front": front, "SiyuanID": SiyuanID}, deck=deck, tags=tags)



from .parser import siyuan as parser
from siyuanhelper import helper as siyuanhelper
from AnkiIn.helper import ankiConnectHelper as ankihelper
from AnkiIn.config import dict as conf
from . import config
from . import config_parser
import datetime
import asyncio
import time


updated_notes = 0
added_notes = 0


async def execute():
    if not ankihelper.check_online():
        print("Anki is not online! Please open Anki and make sure anki-connect is installed.")
        return
    start_time = time.perf_counter()
    config_parser.parse()
    print("Api Token:{}".format(conf["siyuan"]["api_token"]))
    print("Last Sync:{}".format(config.last_sync_time))
    this_sync_time = datetime.datetime.now().strftime(r"%Y%m%d%H%M%S")
    print("Now Sync: {}".format(this_sync_time))
    noteList = await parser.sync(config.last_sync_time)
    for x in noteList:
        if x is None:
            continue
        handle(x)
    print("Added {} notes, updated {} notes.".format(added_notes, updated_notes))
    end_time = time.perf_counter()
    print("Time consumed: {}s".format(end_time - start_time))
    f = open("last_sync_time", "w", encoding="UTF-8")
    f.write(this_sync_time)
    f.close()


def handle(x):
    global updated_notes, added_notes
    old = ankihelper.find_notes("SiyuanID:\"{}\"".format(x.outputfields["SiyuanID"]))
    # print(x.outputfields["SiyuanID"])
    # print(x.SiyuanID)
    if len(old) >= 2:
        print("found SiyuanID[{}] with duplicates in Anki!\n{}".format(
            x["SiyuanID"], old))
    elif len(old) == 1:
        ankihelper.update_note_fields(old[0], x)
        print("updated: SiyuanID[{}] AnkiID[{}]".format(
            x.fields["SiyuanID"], old[0]))
        updated_notes = updated_notes + 1
    else:
        ankihelper.add_note(x)
        added_notes = added_notes + 1


def execute_from_commandline():
    asyncio.run(execute())



from AnkiIn.note import Note
from AnkiIn.model import Model
from AnkiIn.config import dict as conf
from AnkiIn.config import config_updater
from AnkiIn.log import notetype_logger as log
from AnkiIn.notetypes.Cloze import get as super_get
from AnkiIn.notetypes.Cloze import check as super_check
from AnkiIn.notetypes.Cloze import ClozeNote


notetype_name = "SCloze"
if notetype_name not in conf["notetype"]:
    conf["notetype"][notetype_name] = {}
settings = conf["notetype"][notetype_name]

priority = None


def update_siyuan_cloze_config():
    global settings, priority

    priority = settings.get("priority", 22)


config_updater.append((update_siyuan_cloze_config, 10))


def check(lines: list, extra_params={}) -> bool:
    return super_check(lines=lines, extra_params=extra_params)


def get(text: str, deck: str, tags: list, extra_params={}) -> Note:
    return SClozeNote(extra_params["SiyuanID"], super_get(text=text, deck=deck, tags=tags, extra_params=extra_params))


CSS = r""".card {
  font-family: arial;
  font-size: 20px;
  text-align: center;
  color: black;
  background-color: white;
}

.cloze {
 font-weight: bold;
 color: blue;
}
.nightMode .cloze {
 color: lightblue;
}
ul {
display: inline-block;
text-align: left;
}
ol {
display: inline-block;
text-align: left;
}
"""

BACK = r"""{{cloze:Text}}
</br>
<a href="siyuan://blocks/{{SiyuanID}}">Open in SiYuan</a>"""

MODELNAME = "SAnkiLink-Cloze"
MODELID = 1146141120

_model = Model(
    modelId=MODELID,
    modelName=MODELNAME,
    isCloze=1,
    fields=["Text", "Back Extra", "SiyuanID"],
    templates=[
        {
            "Name": "Cloze",
            "Front": "{{cloze:Text}}",
            "Back": BACK
        }
    ],
    css=CSS
)


class SClozeNote(ClozeNote):
    def __init__(self, SiyuanID: str, text: str, deck, tags):
        global _model
        super().__init__(text=text, deck=deck, tags=tags)
        self.model = _model
        self["SiyuanID"] = SiyuanID

    def __init__(self, SiyuanID: str, Note):
        global _model
        self.__dict__.update(Note.__dict__)
        self.model = _model
        self["SiyuanID"] = SiyuanID



from AnkiIn.notetypes.TableCloze import check as super_check
from AnkiIn.notetypes.TableCloze import get as super_get
from AnkiIn.config import dict as conf
from AnkiIn.config import config_updater
from . import SCloze


notetype_name = "STableCloze"
if notetype_name not in conf["notetype"]:
    conf["notetype"][notetype_name] = {}
settings = conf["notetype"][notetype_name]

priority = None


def update_siyuan_table_cloze_config():
    global settings, priority

    priority = settings.get("priority", 15)


config_updater.append((update_siyuan_table_cloze_config, 10))


def check(lines: list, extra_params={}) -> bool:
    return super_check(lines=lines, extra_params=extra_params)


def get(text: str, deck: str, tags: list, extra_params={}):
    return SCloze(extra_params["SiyuanID"], super_get(text=text, deck=deck, tags=tags, extra_params=extra_params))



import argparse
import os.path
from AnkiIn.config import dict as conf
from AnkiIn import config as AnkiIn_config
from . import config
import toml

version_name = "0.0.4"

help_info = {
    "config": "set your config file path."
}

parser = argparse.ArgumentParser(
    description="AnkiSiyuan. Link your data in Siyuan with Anki!")
parser.add_argument("-v", "--version", action="version",
                    version="Anki Siyuan v{}".format(version_name))
parser.add_argument("-c", "--config", metavar="config", default=config.config_path,
                    help=help_info["config"])


def enable_log_file():
    conf["log_config"]["handlers"]["log_file"] = {
        "level": "DEBUG" if config.log_debug else "INFO",
        "formatter": "standard",
        "class": "logging.FileHandler",
        "filename": "log.txt",
        "mode": "a"
    }
    for x in conf["log_config"]["loggers"].keys():
        conf["log_config"]["loggers"][x]["handlers"].append("log_file")


def parse():
    args = parser.parse_args()
    config.config_path = args.config
    conf["siyuan"] = {}
    conf["siyuan"]["api_token"] = ""
    conf["siyuan"]["custom_attr_name"] = "custom-ankilink"
    conf["siyuan"]["assets_replacement"] = "assets"
    try:
        f = open("last_sync_time", "r", encoding="UTF-8")
        config.last_sync_time = f.read()
        f.close()
    except Exception:
        pass
    enable_log_file()
    load_config_file()
    AnkiIn_config.update_config()


def load_config_file():
    if os.path.isfile(config.config_path):
        AnkiIn_config.execute_config(toml.load(config.config_path))
    else:
        print("Warning: Config File {} doesn't exist!".format(config.config_path))



from AnkiIn.helper.formatHelper import remove_suffix
from siyuanhelper.helper import PropertyNotFoundException, do_property_exist_by_id, get_parent_by_id
from siyuanhelper.helper import get_property_by_id, query_sql, get_col_by_id
from siyuanhelper.helper import set_token, set_session
from AnkiIn.parser import markdown
from AnkiIn.notetype_loader import discovered_notetypes
from ..notetypes import SQA, SMQA, SCloze, SListCloze, STableCloze, SWatch
from AnkiIn.config import update_config
from AnkiIn.config import dict as conf
from AnkiIn.config import config_updater
from AnkiIn import config
from AnkiIn.log import parser_logger as logger
import asyncio
import aiohttp


class SyntaxNode:
    def __init__(self, id: str, parent=None, sons=None):
        self.id = id
        self.parent = parent
        self.sons = sons
        if self.sons is None:
            self.sons = []


link = {}
is_added = {}
roots = []
noteList = []
tag_attr_name = "ankilink"
assets_replacement = "assets"


def update_siyuan_parser():
    global tag_attr_name
    global assets_replacement
    if "siyuan" not in conf:
        conf["siyuan"] = {}
    tag_attr_name = conf["siyuan"].get(
        "custom_attr_name", "custom-ankilink")
    assets_replacement = conf["siyuan"].get("assets_replacement", "assets")


discovered_notetypes += [SQA, SMQA, SCloze, SListCloze, STableCloze, SWatch]
config_updater.append((update_siyuan_parser, 5))
update_config()


async def build_tree(now: str):
    # print("visit:{}".format(now))
    # print("build tree")
    # print(get_col_by_id(now, "markdown"))
    now_node = SyntaxNode(now)
    link[now] = now_node
    try:
        if await do_property_exist_by_id(now, tag_attr_name):
            roots.append(now_node)
            return now_node
    except Exception:
        logger.exception("Exception occurred! Invalid Siyuan ID {}".format(now))
        logger.exception(Exception)
    fa_id = await get_parent_by_id(now)
    if fa_id == "":
        return now_node
    # print("son: {} fa:{}".format(now, fa_id))
    fa = link.get(fa_id)
    if fa is None:
        fa = await build_tree(fa_id)
    now_node.parent = fa
    # print("fa {} added son {}".format(fa.id, now_node.id))
    fa.sons.append(now_node)
    # print("fa {} sons:".format(fa.id))
    # print([x.id for x in fa.sons])
    return now_node


async def sync(last_time: str):
    # session = aiohttp.ClientSession()
    # set_session(session)
    set_token(conf["siyuan"].get("api_token", ""))
    session = aiohttp.ClientSession()
    set_session(session)
    link.clear()
    is_added.clear()
    roots.clear()
    noteList.clear()
    all_origin_blocks = await query_sql(
        r"SELECT id FROM blocks where updated>'{}' and type='p'".format(last_time))
    all_blocks = [x["id"] for x in all_origin_blocks]
    # print(all_blocks)
    tasks = []
    for x in all_blocks:
        tasks.append(asyncio.create_task(build_tree(x)))
    await asyncio.tasks.gather(*tasks)
    # print([x.id for x in roots])
    # print([get_col_by_id(x.id,"markdown") for x in roots])
    for x in roots:
        await dfs(x)
    await session.close()
    return noteList


async def dfs(now: SyntaxNode):
    # print("dfs: " + now.id)
    # print([x.id for x in now.sons])
    current_config = None
    config_backup = None
    try:
        current_config = (await get_property_by_id(
            now.id, tag_attr_name))
        config_backup = config.parse_config(current_config)
    except PropertyNotFoundException:
        logger.debug("SiyuanID:{} has no config.".format(now.id))
    # except Exception:
    #    logger.warning(
    #        "An error occurred while parsing config.\nSiyuanID:{}\nProperty:\n{}".format(now.id, current_config))
    if len(now.sons) == 0:
        # print("!!!")
        # print(get_col_by_id(now.id, "markdown"))
        # leaf
        await handle(now)
    else:
        for x in now.sons:
            await dfs(x)
    if config_backup is not None:
        config.execute_config(config_backup)


async def handle(now: SyntaxNode):
    fa = await get_parent_by_id(now.id)
    if (await get_col_by_id(now.id, "type")) == "i" or fa != "" and (await get_col_by_id(fa, "type")) == "i":
        await handle(now.parent)
    else:
        await addNote(now.id)


async def addNote(id):
    if is_added.get(id, False):
        return
    text = spec_format(await get_col_by_id(id, "markdown"))
    note = markdown.get_note(text, extra_params={"SiyuanID": id})
    if note is None:
        return
    noteList.append(note)
    is_added[id] = True


def spec_format(text: str) -> str:
    lines = [x for x in text.splitlines() if x != ""]
    text = ""
    for x in lines:
        if x.startswith(" "):
            # duplicate leading space for lists
            p = 0
            while x[p] == " ":
                p = p + 1
            x = x[:p] + x[:p] + x[p:]
        text = text + x + "\n"
    text = remove_suffix(text, "\n")
    text = text.replace(
        r"(assets/", r"({}/".format(assets_replacement))
    return text



from AnkiIn.note import Note
from AnkiIn.config import dict as conf
from AnkiIn.config import config_updater
from AnkiIn.log import notetype_logger as log
from .SQA import SQANote
from AnkiIn.notetypes.MQA import get as super_get
from AnkiIn.notetypes.MQA import check as super_check


notetype_name = "SMQA"
if notetype_name not in conf["notetype"]:
    conf["notetype"][notetype_name] = {}
settings = conf["notetype"][notetype_name]

priority = None
prefix = None


def update_siyuan_mqa_config():
    global settings, priority, prefix

    priority = settings.get("priority", 13)
    prefix = settings.get("prefix", "!")


config_updater.append((update_siyuan_mqa_config, 10))


def check(lines: list, extra_params={}) -> bool:
    return super_check(lines=lines, extra_params=extra_params)


def get(text: str, deck: str, tags: list, extra_params={}) -> Note:
    return SQANote(extra_params["SiyuanID"], super_get(text=text, deck=deck, tags=tags, extra_params=extra_params))
